class CoordException(Exception):
    """
        Raised when Coord is wrong
    """
    pass
    
class RangeException(Exception):
    """
        Raised when Coord is wrong
    """
    pass

class UnogeneratorException(Exception): 
    pass
  
